@@ -1,4 +1,4 @@
-# $Id: Simple.pm,v 1.18 2004/03/02 08:18:02 grantm Exp $
+# $Id: Simple.pm,v 1.22 2004/11/17 08:36:49 grantm Exp $
 
 package XML::Simple;
 
@@ -53,7 +53,7 @@ use vars qw($VERSION @ISA @EXPORT @EXPORT_OK $PREFERRED_PARSER);
 @ISA               = qw(Exporter);
 @EXPORT            = qw(XMLin XMLout);
 @EXPORT_OK         = qw(xml_in xml_out);
-$VERSION           = '2.11';
+$VERSION           = '2.13';
 $PREFERRED_PARSER  = undef;
 
 my $StrictMode     = 0;
@@ -66,11 +66,12 @@ my %CacheScheme    = (
 my @KnownOptIn     = qw(keyattr keeproot forcecontent contentkey noattr
                         searchpath forcearray cache suppressempty parseropts
                         grouptags nsexpand datahandler varattr variables
-                        normalisespace normalizespace);
+                        normalisespace normalizespace valueattr);
 
 my @KnownOptOut    = qw(keyattr keeproot contentkey noattr
                         rootname xmldecl outputfile noescape suppressempty
-                        grouptags nsexpand handler noindent attrindent nosort);
+                        grouptags nsexpand handler noindent attrindent nosort
+                        valueattr numericescape);
 
 my @DefKeyAttr     = qw(name key id);
 my $DefRootName    = qq(opt);
@@ -330,7 +331,7 @@ sub build_tree_xml_parser {
   if($filename) {
     # $tree = $xp->parsefile($filename);  # Changed due to prob w/mod_perl
     local(*XML_FILE);
-    open(XML_FILE, "<$filename") || croak qq($filename - $!);
+    open(XML_FILE, '<', $filename) || croak qq($filename - $!);
     $tree = $xp->parse(*XML_FILE);
     close(XML_FILE);
   }
@@ -540,8 +541,9 @@ sub XMLout {
     }
     else {
       local(*OUT);
-      open(OUT, ">$self->{opt}->{outputfile}") ||
+      open(OUT, '>', "$self->{opt}->{outputfile}") ||
         croak "open($self->{opt}->{outputfile}): $!";
+      binmode(OUT, ':utf8') if($] >= 5.008);
       print OUT $xml || croak "print: $!";
       close(OUT);
     }
@@ -763,6 +765,15 @@ sub handle_options  {
   }
 
 
+  # Special cleanup for {valueattr} which could be arrayref or hashref
+
+  if(exists($opt->{valueattr})) {
+    if(ref($opt->{valueattr}) eq 'ARRAY') {
+      $opt->{valueattrlist} = {};
+      $opt->{valueattrlist}->{$_} = 1 foreach(@{ delete $opt->{valueattr} });
+    }
+  }
+
   # make sure there's nothing weird in {grouptags}
 
   if($opt->{grouptags} and !UNIVERSAL::isa($opt->{grouptags}, 'HASH')) {
@@ -880,6 +891,16 @@ sub collapse {
   }
 
 
+  # Roll up 'value' attributes (but only if no nested elements)
+
+  if(!@_  and  keys %$attr == 1) {
+    my($k) = keys %$attr;
+    if($self->{opt}->{valueattrlist}  and $self->{opt}->{valueattrlist}->{$k}) {
+      return $attr->{$k};
+    }
+  }
+
+
   # Add any nested elements
 
   my($key, $val);
@@ -1008,6 +1029,19 @@ sub collapse {
     return(undef);
   }
 
+
+  # Roll up named elements with named nested 'value' attributes
+
+  if($self->{opt}->{valueattr}) {
+    while(my($key, $val) = each(%$attr)) {
+      next unless($self->{opt}->{valueattr}->{$key});
+      next unless(UNIVERSAL::isa($val, 'HASH') and (keys %$val == 1));
+      my($k) = keys %$val;
+      next unless($k eq $self->{opt}->{valueattr}->{$key});
+      $attr->{$key} = $val->{$k};
+    }
+  }
+
   return($attr)
 
 }
@@ -1213,13 +1247,13 @@ sub value_to_xml {
 
   my $nl = "\n";
 
+  my $is_root = $indent eq '' ? 1 : 0;   # Warning, dirty hack!
   if($self->{opt}->{noindent}) {
     $indent = '';
     $nl     = '';
   }
 
 
-
   # Convert to XML
   
   if(ref($ref)) {
@@ -1246,12 +1280,12 @@ sub value_to_xml {
   if(UNIVERSAL::isa($ref, 'HASH')      # It is a hash
      and keys %$ref                    # and it's not empty
      and $self->{opt}->{keyattr}       # and folding is enabled
-     and $indent                       # and its not the root element
+     and !$is_root                     # and its not the root element
   ) {
     $ref = $self->hash_to_array($name, $ref);
   }
 
-  
+
   my @result = ();
   my($key, $value);
 
@@ -1263,6 +1297,7 @@ sub value_to_xml {
     # Reintermediate grouped values if applicable
 
     if($self->{opt}->{grouptags}) {
+      $ref = $self->copy_hash($ref);
       while(my($key, $val) = each %$ref) {
         if($self->{opt}->{grouptags}->{$key}) {
           $ref->{$key} = { $self->{opt}->{grouptags}->{$key} => $val };
@@ -1276,7 +1311,7 @@ sub value_to_xml {
     my $nsdecls = '';
     my $default_ns_uri;
     if($self->{nsup}) {
-      $ref = { %$ref };                # Make a copy before we mess with it
+      $ref = $self->copy_hash($ref);
       $self->{nsup}->push_context();
 
       # Look for default namespace declaration first
@@ -1340,6 +1375,7 @@ sub value_to_xml {
         my $value = $ref->{$key};
         next if(substr($key, 0, 1) eq '-');
         if(!defined($value)) {
+          next if $self->{opt}->{suppressempty};
           unless(exists($self->{opt}->{suppressempty})
              and !defined($self->{opt}->{suppressempty})
           ) {
@@ -1349,9 +1385,17 @@ sub value_to_xml {
             $text_content = '';
           }
           else {
-            $value = {};
+            $value = exists($self->{opt}->{suppressempty}) ? {} : '';
           }
         }
+
+        if(!ref($value)  
+           and $self->{opt}->{valueattr}
+           and $self->{opt}->{valueattr}->{$key}
+        ) {
+          $value = { $self->{opt}->{valueattr}->{$key} => $value };
+        }
+
         if(ref($value)  or  $self->{opt}->{noattr}) {
           push @nested,
             $self->value_to_xml($value, $key, "$indent  ");
@@ -1486,7 +1530,24 @@ sub escape_value {
   $data =~ s/>/&gt;/sg;
   $data =~ s/"/&quot;/sg;
 
-  return($data);
+  my $level = $self->{opt}->{numericescape} or return $data;
+
+  return $self->numeric_escape($data, $level);
+}
+
+sub numeric_escape {
+  my($self, $data, $level) = @_;
+
+  use utf8; # required for 5.6
+
+  if($self->{opt}->{numericescape} eq '2') {
+    $data =~ s/([^\x00-\x7F])/'&#' . ord($1) . ';'/gse;
+  }
+  else {
+    $data =~ s/([^\x00-\xFF])/'&#' . ord($1) . ';'/gse;
+  }
+
+  return $data;
 }
 
 
@@ -1515,8 +1576,9 @@ sub hash_to_array {
 
     if(ref($self->{opt}->{keyattr}) eq 'HASH') {
       return($hashref) unless(defined($self->{opt}->{keyattr}->{$parent}));
-      push(@$arrayref, { $self->{opt}->{keyattr}->{$parent}->[0] => $key,
-                         %$value });
+      push @$arrayref, $self->copy_hash(
+        $value, $self->{opt}->{keyattr}->{$parent}->[0] => $key
+      );
     }
     else {
       push(@$arrayref, { $self->{opt}->{keyattr}->[0] => $key, %$value });
@@ -1527,6 +1589,22 @@ sub hash_to_array {
 }
 
 
+##############################################################################
+# Method: copy_hash()
+#
+# Helper routine for hash_to_array().  When unfolding a hash of hashes into
+# an array of hashes, we need to copy the key from the outer hash into the
+# inner hash.  This routine makes a copy of the original hash so we don't
+# destroy the original data structure.  You might wish to override this
+# method if you're using tied hashes and don't want them to get untied.
+#
+
+sub copy_hash {
+  my($self, $orig, @extra) = @_;
+
+  return { @extra, %$orig };
+}
+
 ##############################################################################
 # Methods required for building trees from SAX events
 ##############################################################################
@@ -2098,6 +2176,9 @@ It will return this simpler structure:
     searchpath => [ '/usr/bin', '/usr/local/bin', '/usr/X11/bin' ]
   }
 
+The grouping element (C<< <searchpath> >> in the example) must not contain any
+attributes or elements other than the grouped element.
+
 You can specify multiple 'grouping element' to 'grouped element' mappings in
 the same hashref.  If this option is combined with C<KeyAttr>, the array
 folding will occur first and then the grouped element names will be eliminated.
@@ -2330,12 +2411,29 @@ C<XMLout> will emit XML which is not well formed.
 I<Note: You must have the XML::NamespaceSupport module installed if you want
 C<XMLout()> to translate URIs back to prefixes>.
 
+=head2 NumericEscape => 0 | 1 | 2 I<# out - handy>
+
+Use this option to have 'high' (non-ASCII) characters in your Perl data
+structure converted to numeric entities (eg: &#8364;) in the XML output.  Three
+levels are possible:
+
+0 - default: no numeric escaping (OK if you're writing out UTF8)
+
+1 - only characters above 0xFF are escaped (ie: characters in the 0x80-FF range are not escaped), possibly useful with ISO8859-1 output
+
+2 - all characters above 0x7F are escaped (good for plain ASCII output)
+
 =head2 OutputFile => <file specifier> I<# out - handy>
 
 The default behaviour of C<XMLout()> is to return the XML as a string.  If you
 wish to write the XML to a file, simply supply the filename using the
-'OutputFile' option.  Alternatively, you can supply an IO handle object instead
-of a filename.
+'OutputFile' option.  
+
+This option also accepts an IO handle object - especially useful in Perl 5.8.0 
+and later for output using an encoding other than UTF-8, eg:
+
+  open my $fh, '>:encoding(iso-8859-1)', $path or die "open($path): $!";
+  XMLout($ref, OutputFile => $fh);
 
 =head2 ParserOpts => [ XML::Parser Options ] I<# in - don't use this>
 
@@ -2380,10 +2478,48 @@ string will cause empty elements to be represented as the undefined value or
 the empty string respectively.  The latter two alternatives are a little
 easier to test for in your code than a hash with no keys.
 
-The option also controls what C<XMLout()> does with undefined values.
-Setting the option to undef causes undefined values to be output as
-empty elements (rather than empty attributes), it also suppresses the
-generation of warnings about undefined values.
+The option also controls what C<XMLout()> does with undefined values.  Setting
+the option to undef causes undefined values to be output as empty elements
+(rather than empty attributes), it also suppresses the generation of warnings
+about undefined values.  Setting the option to a true value (eg: 1) causes
+undefined values to be skipped altogether on output.
+
+=head2 ValueAttr => [ names ] I<# in - handy>
+
+Use this option to deal elements which always have a single attribute and no
+content.  Eg:
+
+  <opt>
+    <colour value="red" />
+    <size   value="XXL" />
+  </opt>
+
+Setting C<< ValueAttr => [ 'value' ] >> will cause the above XML to parse to:
+
+  {
+    colour => 'red',
+    size   => 'XXL'
+  }
+
+instead of this (the default):
+
+  {
+    colour => { value => 'red' },
+    size   => { value => 'XXL' }
+  }
+
+Note: This form of the ValueAttr option is not compatible with C<XMLout()> -
+since the attribute name is discarded at parse time, the original XML cannot be
+reconstructed.
+
+=head2 ValueAttr => { element => attribute, ... } I<# in+out - handy>
+
+This (preferred) form of the ValueAttr option requires you to specify both
+the element and the attribute names.  This is not only safer, it also allows
+the original XML to be reconstructed by C<XMLout()>.
+
+Note: You probably don't want to use this option and the NoAttr option at the
+same time.
 
 =head2 Variables => { name => value } I<# in - handy>
 
